from sigma.processing.postprocessing import QueryPostprocessingTransformation
from sigma.rule import SigmaRule


class PrependMetadataPostprocessingTransformation(QueryPostprocessingTransformation):
    def apply(self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: str) -> str:  # type: ignore # noqa: F821
        event_types = rule.custom_attributes.get("event_types", set())
        if pipeline.state.get("output_format", "default") == "yara_l":
            metadata_eventtype = " OR ".join(
                [f'$event1.metadata.event_type = "{event_type}"' for event_type in event_types]
            )
            return f"{metadata_eventtype}\n\n{query}"
        else:
            metadata_eventtype = " OR ".join([f'metadata.event_type = "{event_type}"' for event_type in event_types])
            return f"({metadata_eventtype}) AND ({query})"


class YaraLPostprocessingTransformation(QueryPostprocessingTransformation):
    def apply(self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: str) -> str:  # type: ignore # noqa: F821
        # Split the query into lines and remove any leading/trailing whitespace
        query_lines = [line.strip() for line in query.split("\n") if line.strip()]

        # Join the lines with proper indentation
        indented_query = "\n    ".join(query_lines)

        return f"""
rule {rule.title.lower().replace(" ", "_")} {{
  meta:
    id = "{rule.id}"
    title = "{rule.title}"
    description = "{rule.description}"
    author = "{rule.author}"
    reference = "{", ".join(rule.references)}"
    date = "{rule.date}"
    tags = "{", ".join(str(t) for t in rule.tags)}"
    severity = "{rule.level}"
    falsepositives = "{", ".join(rule.falsepositives) if rule.falsepositives else "Unknown"}" 

  events:
    {indented_query}
    
  conditions: 
    $event1
}}
    """



import re
from dataclasses import dataclass
from typing import Iterable, List, Optional, Union

from sigma.correlations import SigmaCorrelationRule
from sigma.processing.pipeline import ProcessingPipeline
from sigma.processing.transformations import (
    DetectionItemTransformation,
    FieldMappingTransformation,
    Transformation,
    ValueTransformation,
)
from sigma.rule import SigmaDetection, SigmaDetectionItem, SigmaRule
from sigma.types import SigmaString, SigmaType

from .errors import InvalidUDMFieldError
from .mappings import enum_mappings, get_field_mappings_by_event_type
from .utils import determine_event_type_event_id, determine_event_type_logsource
from .validators import is_valid_udm_field


@dataclass
class PrependEventVariableTransformation(FieldMappingTransformation):
    """Prepend event variable to every field name for YARA_L output format/pipeline"""

    mapping = {}

    def get_mapping(self, field: str) -> Union[None, str, List[str]]:
        return f"$event1.{field}"


@dataclass
class SetPrependMetadataTransformation(Transformation):
    prepend_metadata: bool

    def apply(self, pipeline: ProcessingPipeline, rule: SigmaRule) -> None:
        self.processing_item_applied(rule)
        if pipeline.state.get("prepend_metadata", None) is None:
            pipeline.state["prepend_metadata"] = self.prepend_metadata


@dataclass
class RemoveHashAlgoFromValueTransformation(ValueTransformation):
    def apply_value(self, field: str, val: SigmaType) -> SigmaType:
        return SigmaString(val.to_plain().split("|")[-1])


@dataclass
class ConvertEnumValueTransformation(ValueTransformation):
    """
    Convert the value of a field to an enum value, with modified ValueTransformation.
    """

    def apply_detection_item(self, detection_item: SigmaDetectionItem):
        """Call apply_value for each value and integrate results into value list."""
        results = []
        modified = False
        if detection_item.field in enum_mappings:
            for value in detection_item.value:
                if self.value_types is None or isinstance(
                    value, self.value_types
                ):  # run replacement if no type annotation is defined or matching to type of value
                    res = self.apply_value(detection_item.field, value)
                    if res is None:  # no value returned: drop value
                        results.append(value)
                    elif isinstance(res, Iterable) and not isinstance(res, SigmaType):
                        results.extend(res)
                        modified = True
                    else:
                        results.append(res)
                        modified = True
                else:  # pass original value if type doesn't matches to apply_value argument type annotation
                    results.append(value)
            if modified:
                detection_item.value = results
                self.processing_item_applied(detection_item)

    def apply_value(self, field: str, val: SigmaType) -> SigmaType:
        return SigmaString(enum_mappings.get(field, {}).get(val.to_plain(), None)) or val


@dataclass
class EnsureValidUDMFieldsTransformation(DetectionItemTransformation):
    """
    Ensure that all fields in the detection item are valid UDM fields.
    """

    udm_schema: dict

    def apply_detection_item(self, detection_item: SigmaDetectionItem) -> None:
        if detection_item.field:
            if not is_valid_udm_field(detection_item.field, self.udm_schema):
                raise InvalidUDMFieldError(f"Field {detection_item.field} is not a valid UDM field")


@dataclass
class SetRuleEventTypeFromLogsourceTransformation(Transformation):
    """
    Sets the `event_types` custom attribute on a rule, that can be used by the processing pipeline and backend during processing.
    `event_types` is a set of event types that the rule is related to.
    If the event_types custom attribute is already set from a previous pipeline, we ensure it is a set
    """

    def apply(
        self,
        pipeline: ProcessingPipeline,
        rule: Union[SigmaRule, SigmaCorrelationRule],
    ) -> None:
        self._pipeline: "sigma.processing.pipeline.ProcessingPipeline" = pipeline  # noqa: F821 # type: ignore
        if rule.custom_attributes.get("event_types", None):
            self.processing_item_applied(rule)
            if isinstance(rule.custom_attributes["event_types"], list) or isinstance(rule.custom_attributes["event_types"], str):
                rule.custom_attributes["event_types"] = set([rule.custom_attributes["event_types"]])
        else:
            rule.custom_attributes["event_types"] = set()
            if event_type := determine_event_type_logsource(rule):
                rule.custom_attributes["event_types"].add(event_type)
                self.processing_item_applied(rule)


@dataclass
class SetRuleEventTypeFromEventIDTransformation(DetectionItemTransformation):
    """Iterates through "selection" detection sections of a rule and sets the event_type custom attribute on the rule if the EventID field is present."""

    def apply_detection_item(self, detection_item: SigmaDetectionItem) -> Optional[str]:
        """Apply transformation on detection item. We need to set the event_type custom attribute on the rule, so we return the event_type string."""
        event_types = set()
        if detection_item.field == "EventID":
            for value in detection_item.value:
                if event_type := determine_event_type_event_id(str(value.to_plain())):
                    event_types.add(event_type)
        if event_types:
            return event_types

    def apply_detection(self, detection: SigmaDetection) -> Optional[str]:
        """Apply transformation on detection. We need to set the event_type custom attribute on the rule, so we return the event_type string."""
        for i, detection_item in enumerate(detection.detection_items):
            if isinstance(detection_item, SigmaDetection):  # recurse into nested detection items
                self.apply_detection(detection_item)
            else:
                if (
                    self.processing_item is None
                    or self.processing_item.match_detection_item(self._pipeline, detection_item)
                ) and (r := self.apply_detection_item(detection_item)) is not None:
                    self.processing_item_applied(detection.detection_items[i])
                    return r

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule  # noqa: F821 # type: ignore
    ) -> None:
        super().apply(pipeline, rule)
        if isinstance(rule, SigmaRule):
            for section_title, detection in rule.detection.detections.items():
                if re.match(r"^sel.*", section_title):
                    if (r := self.apply_detection(detection)) is not None:
                        rule.custom_attributes["event_types"].update(r)


class EventTypeFieldMappingTransformation(FieldMappingTransformation):
    """
    Dynamically sets the mapping dictionary based on the Sigma rule's custom attribute 'event_types'.
    """

    def __init__(self):
        super().__init__({})  # Initialize parent class with an empty mapping for now

    def apply(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",  # noqa: F821 # type: ignore
        rule: Union[SigmaRule, SigmaCorrelationRule],  # noqa: F821 # type: ignore
    ) -> None:
        """Apply dynamic mapping before the field name transformations."""
        self.set_event_type_mapping(rule)  # Dynamically update the mapping
        super().apply(pipeline, rule)  # Call parent method to continue the transformation process

    def set_event_type_mapping(self, rule: SigmaRule):
        """
        Set the mapping dynamically based on the rule's custom attribute 'event_types'.
        Update the mapping dict to ensure the first event type field mappings
        override the last ones in the event the same field is present in multiple event types mappings
        """
        event_types = rule.custom_attributes.get("event_types", set())
        if event_types:
            # Convert set to list and reverse it
            event_types_list = list(event_types)
            event_types_list.reverse()

            for event_type in event_types_list:
                if event_type_mappings := get_field_mappings_by_event_type(event_type):
                    self.mapping.update(event_type_mappings)



import json
from importlib import resources

from sigma.processing.conditions import (
    DetectionItemProcessingItemAppliedCondition,
    IncludeFieldCondition,
    RuleProcessingItemAppliedCondition,
    RuleProcessingStateCondition,
)
from sigma.processing.pipeline import ProcessingItem, ProcessingPipeline, QueryPostprocessingItem
from sigma.processing.transformations import (
    FieldMappingTransformation,
)

from .mappings import enum_mappings, get_field_mappings
from .postprocessing import PrependMetadataPostprocessingTransformation
from .transformations import (
    ConvertEnumValueTransformation,
    EnsureValidUDMFieldsTransformation,
    EventTypeFieldMappingTransformation,
    RemoveHashAlgoFromValueTransformation,
    SetPrependMetadataTransformation,
    SetRuleEventTypeFromEventIDTransformation,
    SetRuleEventTypeFromLogsourceTransformation,
)

# LOAD UDM SCHEMA
udm_schema = json.loads(resources.read_text("sigma.pipelines.secops", "udm_field_schema.json"))

# PROCESSING ITEMS

## SET PIPELINE STATE FOR POSTPROCESSING


def set_prepend_metadata_proc_item(prepend_metadata: bool = True) -> ProcessingItem:
    return ProcessingItem(
        identifier="secops_set_prepend_metadata",
        transformation=SetPrependMetadataTransformation(prepend_metadata),
    )


## SET EVENT TYPE IN RULE CUSTOM ATTRIBUTE

set_event_type_proc_items = [
    ProcessingItem(
        identifier="secops_set_event_type_from_logsource",
        transformation=SetRuleEventTypeFromLogsourceTransformation(),
    ),
    ProcessingItem(
        identifier="secops_set_event_type_from_event_id",
        transformation=SetRuleEventTypeFromEventIDTransformation(),
        field_name_conditions=[
            IncludeFieldCondition(["EventID"]),
        ],
        rule_conditions=[RuleProcessingItemAppliedCondition("secops_set_event_type_from_logsource")],
        rule_condition_negation=True,  # If we can set the event type from the logsource, we don't need to set it from any EventIDs present in selection items
    ),
]


## FIELD MAPPINGS

event_type_field_mapping_proc_item = ProcessingItem(
    identifier="secops_event_type_field_mappings",
    transformation=EventTypeFieldMappingTransformation(),
)

# If field has not been mapped by event_type_field_mapping_proc_item, map using common_field_mappings
common_field_mappings_proc_item = ProcessingItem(
    identifier="secops_common_field_mappings",
    transformation=FieldMappingTransformation(get_field_mappings().get("common")),
    detection_item_conditions=[DetectionItemProcessingItemAppliedCondition("secops_event_type_field_mappings")],
    detection_item_condition_linking=any,
    detection_item_condition_negation=True,
)

## CONVERT ENUM VALUES IF POSSIBLE

convert_enum_values_proc_item = ProcessingItem(
    identifier="secops_convert_enum_values",
    transformation=ConvertEnumValueTransformation(),
    field_name_conditions=[
        IncludeFieldCondition(list(enum_mappings.keys())),
    ],
)

# UDM VALIDATION
## ENSURE ALL FIELDS ARE VALID UDM FIELDS

udm_validation_proc_item = ProcessingItem(
    identifier="secops_udm_validation",
    transformation=EnsureValidUDMFieldsTransformation(udm_schema),
)

## REMOVE HASH ALGORITHM FROM HASHES FIELD
### MUST OCCUR AFTER FIELD RENAMES

remove_hash_algo_from_hashes_proc_item = ProcessingItem(
    identifier="secops_remove_hash_algo_from_hashes",
    transformation=RemoveHashAlgoFromValueTransformation(),
    field_name_conditions=[
        IncludeFieldCondition(["hash"]),
    ],
)

# POSTPROCESSING
## ADD METADATA TO QUERY

prepend_metadata_postprocessing_item = QueryPostprocessingItem(
    identifier="secops_prepend_metadata_postprocessing",
    transformation=PrependMetadataPostprocessingTransformation(),
    rule_conditions=[RuleProcessingStateCondition(key="prepend_metadata", val=True)],
)


def secops_udm_pipeline(prepend_metadata: bool = True) -> ProcessingPipeline:
    return ProcessingPipeline(
        name="Google SecOps UDM Pipeline",
        priority=20,
        items=[
            set_prepend_metadata_proc_item(prepend_metadata),
            *set_event_type_proc_items,
            event_type_field_mapping_proc_item,
            common_field_mappings_proc_item,
            udm_validation_proc_item,
            convert_enum_values_proc_item,
            remove_hash_algo_from_hashes_proc_item,
        ],
        postprocessing_items=[prepend_metadata_postprocessing_item],
        allowed_backends=["secops"],
    )



from sigma.processing.pipeline import ProcessingPipeline, ProcessingItem, QueryPostprocessingItem
from .transformations import PrependEventVariableTransformation
from .postprocessing import YaraLPostprocessingTransformation


prepend_event_variable_transformation = ProcessingItem(
    identifier="yara_l_prepend_event_variable",
    transformation=PrependEventVariableTransformation(mapping={}),
)

output_format_postprocessing_item = QueryPostprocessingItem(
    identifier="yara_l_output_format_postprocessing",
    transformation=YaraLPostprocessingTransformation(),
)


def yara_l_pipeline() -> ProcessingPipeline:
    """Google SecOps backend format pipeline for YARA-L 2.0 output format
    Not for use as a general purpose pipeline, use the secops pipeline instead.
    """

    return ProcessingPipeline(
        name="Google SecOps YARA-L 2.0 Output Format Pipeline",
        priority=60,
        items=[prepend_event_variable_transformation],
        postprocessing_items=[output_format_postprocessing_item],
        allowed_backends=["secops"],
    )



from functools import lru_cache
from typing import Dict
from .utils import get_field_mapping_type

# TODO: EventType field in sysmon logs, ensure value is valid for metadata.event_type
# TODO: Add mappings/transformations for LogonType to Authentication.Mechanism enum conversion for logon events


@lru_cache(maxsize=128)
def get_field_mappings() -> Dict[str, Dict[str, str]]:
    return {
        "common": {
            "AccessMask": "target.process.access_mask",
            "User": "user",
            "Image": "file_path",
            "ParentImage": "file_path",
            "ProcessId": "process_id",
            "ParentProcessId": "process_id",
            "SourceHostname": "principal.hostname",
            "DestinationHostname": "target.hostname",
            "EventID": "metadata.product_event_type",
            "Provider_Name": "metadata.product_name",
            "ServiceName": "target.process.file.names",
            "ServiceFileName": "file_path",
            "AccountName": "user",
            "SubjectUserName": "principal.user.userid",
            "SubjectDomainName": "domain",
            "TargetUserName": "target.user.userid",
            "TargetDomainName": "domain",
            "IpAddress": "ip",
            "IpPort": "principal.port",
            "WorkstationName": "hostname",
            "Hostname": "hostname",
            "ComputerName": "hostname",
            "Hashes": "hash",
        },
        "process": {
            "CommandLine": "target.process.command_line",
            "CurrentDirectory": "target.process.file.full_path",
            "Image": "target.process.file.full_path",
            "OriginalFileName": "target.process.file.names",
            "ParentImage": "principal.process.file.full_path",
            "ParentCommandLine": "principal.process.command_line",
            "ProcessGuid": "target.process.product_specific_process_id",
            "ProcessId": "target.process.pid",
            "ParentProcessImage": "principal.process.file.full_path",
            "ParentProcessCommandLine": "principal.process.command_line",
            "ParentProcessId": "principal.process.pid",
            "ParentProcessGuid": "principal.process.product_specific_process_id",
            "ParentUser": "principal.user.userid",
            "IntegrityLevel": "target.process.integrity_level_rid",
            "User": "target.user.userid",
        },
        "network": {
            "SourceIp": "principal.ip",
            "DestinationIp": "target.ip",
            "SourcePort": "principal.port",
            "DestinationPort": "target.port",
            "Protocol": "network.ip_protocol",
            "Image": "principal.process.file.full_path",
            "Initiated": "network.direction",
            "User": "principal.user.userid",
            "DestinationHostname": "target.hostname",
            "SourceHostname": "principal.hostname",
            "QueryName": "network.dns.questions.name",
            "QueryResults": "network.dns.answers.data",
            "QueryStatus": "network.dns.response_code",
        },
        "file": {
            "TargetFilename": "target.file.names",
            "Image": "target.process.file.full_path",
            "ObjectName": "target.file.full_path",
            "OldName": "target.file.names",
            "NewName": "target.file.names",
            "OriginalFileName": "target.file.names",
        },
        "authentication": {
            "TargetUserName": "target.user.userid",
            "SubjectUserName": "principal.user.user_display_name",
            "TargetOutboundUserName": "target.user.userid",
            "TargetUserSid": "target.user.windows_sid",
            "TargetServerName": "target.hostname",
            "WorkstationName": "principal.hostname",
            "IpAddress": "principal.ip",
            "IpPort": "principal.port",
            "LogonGuid": "principal.user.product_object_id",
        },
        "registry": {
            "TargetObject": "target.registry.registry_key",
            "Details": "target.registry.registry_value_data",
            "EventType": "metadata.event_type",
            "Image": "principal.process.file.full_path",
            "ProcessId": "principal.process.pid",
            "User": "principal.user.userid",
            "ObjectName": "target.registry.registry_key",
            "ObjectValueName": "target.registry.registry_value_name",
            "NewName": "target.registry.registry_key",
        },
        "dns": {
            "QueryName": "network.dns.questions.name",
            "QueryResults": "network.dns.answers.data",
            "QueryStatus": "network.dns.response_code",
            "record_type": "network.dns.questions.type",
            "answers": "network.dns.answers.name",
        },
    }


def get_field_mappings_by_event_type(metadata_event_type: str) -> Dict[str, str]:
    """
    Get the field mappings for a given event type

    Args:
        metadata_event_type (str): The event type to get the field mappings for

    Returns:
        Dict[str, str]: The field mappings for the given event type
    """

    event_type = get_field_mapping_type(metadata_event_type)
    if not event_type:
        return {}

    all_mappings = get_field_mappings()
    return all_mappings.get(event_type, {})


enum_mappings = {
    "network.direction": {  # From Initiated sysmon field
        "Inbound": "INBOUND",
        "Outbound": "OUTBOUND",
        "Broadcast": "BROADCAST",
        "true": "OUTBOUND",
        "false": "INBOUND",
    }
}
